#include <windows.h>

#include "zlib.h"
#include "zconf.h"

#pragma comment(lib, "zlib.lib")

#define PK0304				0x04034B50
#define PK0102				0x02014B50
#define CRC32_DEFAULT		0xFFFFFFFF
#define CRC32_POLY			0xEDB88320

#pragma pack(push, 2)
struct ZipHeader
{
	unsigned long Signature;
	unsigned short NeedVersion;
	unsigned short Option;
	unsigned short CompressMethod;
	unsigned short FileTime;
	unsigned short FileDate;
	unsigned long Crc32;
	unsigned long CompressSize;
	unsigned long UncompressSize;
	unsigned short FileNameLength;
	unsigned short ExtraLength;
};

struct ZipCentralHeader
{
	unsigned long Signature;
	unsigned short MadeVersion;
	unsigned short NeedVersion;
	unsigned short Option;
	unsigned short CompressMethod;
	unsigned short FileTime;
	unsigned short FileDate;
	unsigned long Crc32;
	unsigned long CompressSize;
	unsigned long UncompressSize;
	unsigned short FileNameLength;
	unsigned short ExtraLength;
	unsigned short CommentLength;
	unsigned short DiskNumber;
	unsigned short DataAttribute;
	unsigned long FileAttribute;
	unsigned long HeaderPosition;
};
#pragma pack(pop)

static unsigned long table[256];

void InitCRC32()
{
	unsigned long u, i, j;

	for(i = 0; i < 256; i++) {
		u = i;
		for(j = 0; j < 8; j++) {
			if(u & 0x01) {
				u = (u >> 1) ^ CRC32_POLY;
			} else {
				u >>= 1;
			}
		}
		table[i] = u;
	}
}

unsigned long CalcCRC32(unsigned char *pt, unsigned long len, unsigned long crc32)
{
	unsigned long no;
	unsigned long result = crc32;

	for(no = 0 ; no < len; no++) {
		result = (result >> 8) ^ table[*pt++ ^ (result & 0xFF)];
	}
	return result;
}

#ifdef _UNICODE
static void extract_directory(wchar_t *dir, const wchar_t *path)
{
	wchar_t *pt;

	pt = NULL;
	while((*dir = *path) != L'\0') {
		if(*path == L'\\') {
			pt = dir;
		}
		path++;
		dir++;
	}
	if(pt != NULL) {
		*pt = L'\0';
	}
}

static void force_directories(const wchar_t *path)
{
	int pos;
	wchar_t dir[MAX_PATH];
	wchar_t *pt = const_cast<wchar_t *>(path);

	pos = 0;
	while(*pt != L'\0') {
		if(*pt == L'\\' && pos > 2) {
			dir[pos] = L'\0';
			CreateDirectory(dir, NULL);
		}
		dir[pos++] = *pt++;
	}
	CreateDirectory(path, NULL);
}

static void change_directory_char(wchar_t *path)
{
	while(*path != L'\0') {
		if(*path == L'/') {
			*path = L'\\';
		}
		path++;
	}
}

#else
static int iskanji(unsigned char ch)
{
	if(((ch >= 0x81) && (ch <= 0x9f)) || ((ch >= 0xe0) && (ch <= 0xfc))){
		return TRUE;
	}
	return FALSE;
}
static void extract_directory(char *dir, const char *path)
{
	int flag;
	char *pt;

	pt = NULL;
	flag = FALSE;
	while((*dir = *path) != '\0') {
		if(flag) {
			flag = FALSE;
		} else {
			if(iskanji((unsigned char)*path)) {
				flag = TRUE;
			} else if(*path == '\\') {
				pt = dir;
			}
		}
		path++;
		dir++;
	}
	if(pt != NULL) {
		*pt = '\0';
	}
}

static void force_directories(const char *path)
{
	int pos;
	char dir[MAX_PATH];
	char *pt = const_cast<char *>(path);
	int flag;

	flag = FALSE;
	pos = 0;
	while(*pt != '\0') {
		if(flag) {
			flag = FALSE;
		} else {
			if(iskanji((unsigned char)*path)) {
				flag = TRUE;
			} else if(*pt == '\\' && pos > 2) {
				dir[pos] = '\0';
				CreateDirectory(dir, NULL);
			}
		}
		dir[pos++] = *pt++;
	}
	CreateDirectory(path, NULL);
}

static void change_directory_char(char *path)
{
	while(*path != '\0') {
		if(*path == '/') {
			*path = '\\';
		}
		path++;
	}
}
#endif

void GetZipFileTime(FILETIME &ft, unsigned short filedate, unsigned short filetime)
{
	SYSTEMTIME st = { 0 };
	FILETIME lft;

	st.wYear = ((filedate >> 9) & 0x7f) + 1980;
	st.wMonth = (filedate >> 5) & 0x0f;
	st.wDay = filedate & 0x1f;
	st.wHour = ((filetime >> 11) & 0x1f);
	st.wMinute = (filetime >> 5) & 0x3f;
	st.wSecond = (filetime & 0x1f) * 2;

	SystemTimeToFileTime(&st, &lft);
	LocalFileTimeToFileTime(&lft, &ft);
}

#define TEMP_MAX	100000

int unzip_memory(char *buff, const TCHAR *out_path)
{
	TCHAR path[MAX_PATH];
	TCHAR dir[MAX_PATH];
	TCHAR filename[MAX_PATH];
	DWORD len;
	HANDLE wfp;
	FILETIME ft;
	unsigned long read_length, rest_length, out_length, crc32;
	unsigned char inflate_buff[TEMP_MAX];
	ZipHeader *header;
	ZipCentralHeader *cheader;
	z_stream zs = { 0 };
	char *top = buff;
	int count = 0;

	InitCRC32();
	inflateInit2(&zs, -15);
	while(1) {
		header = (ZipHeader *)buff;
		// wb_^Cv̔
		if(header->Signature != PK0304) {
			break;
		}
		buff += sizeof(ZipHeader) + header->FileNameLength + header->ExtraLength + header->CompressSize;
	}
	while(1) {
		cheader = (ZipCentralHeader *)buff;
		// wb_^Cv̔
		if(cheader->Signature != PK0102) {
			break;
		}
#ifdef _UNICODE
		MultiByteToWideChar(CP_ACP, 0, buff + sizeof(ZipCentralHeader), cheader->FileNameLength, filename, MAX_PATH);
#else
		lstrcpyn(filename, buff + sizeof(ZipCentralHeader), MAX_PATH);
#endif
		filename[cheader->FileNameLength] = '\0';
		buff += sizeof(ZipCentralHeader) + cheader->FileNameLength + cheader->ExtraLength + cheader->CommentLength;

		char *pt = top + cheader->HeaderPosition;
		header = (ZipHeader *)pt;
		pt += sizeof(ZipHeader) + header->FileNameLength + header->ExtraLength;
		change_directory_char(filename);
		wsprintf(path, "%s\\%s", out_path, filename);
		if(cheader->FileAttribute & FILE_ATTRIBUTE_DIRECTORY) {
			force_directories(path);
			count++;
		} else {
			extract_directory(dir, path);
			force_directories(dir);

			if((wfp = CreateFile(path, GENERIC_WRITE, 0, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL, 0)) == INVALID_HANDLE_VALUE) {
				wfp = CreateFile(path, GENERIC_WRITE, 0, NULL, OPEN_ALWAYS | TRUNCATE_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
			}
			if(wfp != INVALID_HANDLE_VALUE) {
				rest_length = cheader->CompressSize;
				if(cheader->CompressMethod == 8) {
					inflateReset(&zs);

					crc32 = CRC32_DEFAULT;
					while(rest_length > 0) {
						read_length = (rest_length > TEMP_MAX) ? TEMP_MAX : rest_length;
						rest_length -= read_length;

						zs.avail_in = read_length;
						zs.next_in = (unsigned char *)pt;
						do {
							zs.avail_out = TEMP_MAX;
							zs.next_out = inflate_buff;
							inflate(&zs, Z_NO_FLUSH);

							out_length = TEMP_MAX - zs.avail_out;

							crc32 = CalcCRC32(inflate_buff, out_length, crc32);

							WriteFile(wfp, inflate_buff, out_length, &len, NULL);
						} while(zs.avail_out == 0);
						pt += read_length;
					}
					if(cheader->Crc32 == ~crc32) {
						count++;
					}
				} else if(cheader->CompressMethod == 0) {
					WriteFile(wfp, pt, cheader->UncompressSize, &len, NULL);
					count++;
				}
				GetZipFileTime(ft, cheader->FileDate, cheader->FileTime);
				SetFileTime(wfp, &ft, NULL, &ft);
				CloseHandle(wfp);
			}
		}
	}
	inflateEnd(&zs);

	return count;
}


