//
//	cpfShell C
//
//	2001/10/18  by takapyu
//	2010/7/26 update
//
#pragma warning (disable: 4786)

#include <windows.h>
#include <windowsx.h>
#include <winuser.h>
#include <shlobj.h>
#include <initguid.h>
#include <shlguid.h>
#include <ctype.h>
#include <htmlhelp.h>
#include <sstream>
#include <iomanip>

#include "resource.h"

#include "cpfshell.h"
#include "cpfsub.h"

#pragma comment(lib, "htmlhelp.lib")

using namespace std;

const char *RegFlagList[] = {
	"flag_ext",
	"flag_folder"
};
const char *RegModeList[] = {
	"mode_ext",
	"mode_folder"
};
const char *RegUnderList[] = {
	"under_ext",
	"under_folder"
};
const char *RegAddList[] = {
	"add_ext",
	"add_folder"
};
const char *RegSameList[] = {
	"same_ext",
	"same_folder"
};
const char *RegSameFlagList[] = {
	"same_flag_ext",
	"same_flag_folder"
};
enum {
	optionExtMode,
	optionExtSelect,
	optionExtUnder,
	optionExtAdd,
	optionExtSame,
	optionFolderMode,
	optionFolderUnder,
	optionFolderAdd,
	optionFolderSame,

	optionMax,
};

#define CPFSHELL_REG		"Software\\nanshiki\\cpfShell"
#define REG_SELECT			"select"
#define REG_EXP				"regexp"

#define HISTORY_SIZE	10	// 

#define HLP_IDH_MAIN	4
#define HLP_IDH_OPTION	6

#define WM_PROGRESS_COPY	(WM_APP+0x0001)

#ifndef ICON_SMALL
#define ICON_SMALL	0
#endif
#ifndef ICON_BIG
#define ICON_BIG	1
#endif

const char *preferred_dropeffect = "Preferred DropEffect";

BOOL APIENTRY InputFunc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
BOOL APIENTRY CopyFunc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
BOOL APIENTRY OptionFunc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
BOOL APIENTRY LogFunc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);

// NXID {76F7FE62-BCEE-11d5-96CD-444553540000}
DEFINE_GUID(CLSID_ShellExtension, 0x76f7fe62L, 0xbcee, 0x11d5, 0x96, 0xcd, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00);

UINT ReferenceCount = 0;
HINSTANCE hInstanceDll = NULL;
LPCSHELLEXT pShellExt;			// VFgIuWFNg

//
//	ւ Shell Extension ̃Tv̂܂܂肵āB
//
extern "C" int APIENTRY
DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
	if(dwReason == DLL_PROCESS_ATTACH) {
		hInstanceDll = hInstance;
	}
	return TRUE;
}

STDAPI DllCanUnloadNow(void)
{
	return (ReferenceCount == 0 ? S_OK : S_FALSE);
}

string GetString(int id)
{
	vector<char> buff;
	buff.reserve(MAX_PATH + 1);
	int len = LoadString(hInstanceDll, id, &buff[0], MAX_PATH);
	string str(&buff[0], len);
	return str;
}

STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID *ppvOut)
{
	*ppvOut = NULL;
	if(IsEqualIID(rclsid, CLSID_ShellExtension)) {
		CShellExtClassFactory *pcf = new CShellExtClassFactory;

		return pcf->QueryInterface(riid, ppvOut);
	}
	return CLASS_E_CLASSNOTAVAILABLE;
}

CShellExtClassFactory::CShellExtClassFactory()
{
	m_cRef = 0L;
	ReferenceCount++;
}

CShellExtClassFactory::~CShellExtClassFactory()
{
	ReferenceCount--;
}

STDMETHODIMP CShellExtClassFactory::QueryInterface(REFIID riid, LPVOID FAR *ppv)
{
	*ppv = NULL;
	if(IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IClassFactory)) {
		*ppv = (LPCLASSFACTORY)this;
		AddRef();
		return NOERROR;
	}
	return E_NOINTERFACE;
}

STDMETHODIMP_(ULONG) CShellExtClassFactory::AddRef()
{
	return ++m_cRef;
}

STDMETHODIMP_(ULONG) CShellExtClassFactory::Release()
{
	if(--m_cRef) {
		return m_cRef;
	}
	delete this;
	return 0L;
}

STDMETHODIMP CShellExtClassFactory::CreateInstance(LPUNKNOWN pUnkOuter, REFIID riid, LPVOID *ppvObj)
{
	*ppvObj = NULL;
	if(pUnkOuter) {
		return CLASS_E_NOAGGREGATION;
	}
	// VFgIuWFNg쐬
	pShellExt = new CShellExt();
	if(pShellExt == NULL) {
		return E_OUTOFMEMORY;
	}
	return pShellExt->QueryInterface(riid, ppvObj);
}

STDMETHODIMP CShellExtClassFactory::LockServer(BOOL fLock)
{
	return NOERROR;
}

CShellExt::CShellExt()
{
	m_cRef = 0L;
	m_pDataObj = NULL;

	ReferenceCount++;
}

CShellExt::~CShellExt()
{
	if(m_pDataObj) {
		m_pDataObj->Release();
	}
	ReferenceCount--;
}

STDMETHODIMP CShellExt::QueryInterface(REFIID riid, LPVOID FAR *ppv)
{
	*ppv = NULL;
	if(IsEqualIID(riid, IID_IShellExtInit) || IsEqualIID(riid, IID_IUnknown)) {
		*ppv = (LPSHELLEXTINIT)this;
	} else if(IsEqualIID(riid, IID_IContextMenu)) {
		*ppv = (LPCONTEXTMENU)this;
	}
	if(*ppv) {
		AddRef();
		return NOERROR;
	}
	return E_NOINTERFACE;
}

STDMETHODIMP_(ULONG) CShellExt::AddRef()
{
	return ++m_cRef;
}

STDMETHODIMP_(ULONG) CShellExt::Release()
{
	if(--m_cRef)
		return m_cRef;
	delete this;
	return 0L;
}

STDMETHODIMP CShellExt::Initialize(LPCITEMIDLIST pIDFolder, LPDATAOBJECT pDataObj, HKEY hRegKey)
{
	FORMATETC fmt = { CF_HDROP, NULL, DVASPECT_CONTENT, -1, TYMED_HGLOBAL };
	STGMEDIUM stg = { TYMED_HGLOBAL };
	HDROP hDrop;
	UINT uNumFiles;
	char name[MAX_PATH];

	if(m_pDataObj) {
		m_pDataObj->Release();
	}
	if(pDataObj) {
		m_pDataObj = pDataObj;
		pDataObj->AddRef();
	}

	if(FAILED(pDataObj->GetData(&fmt, &stg))) {
		return E_INVALIDARG;
	}
	hDrop = (HDROP)GlobalLock(stg.hGlobal);
	// t@C𓾂
	uNumFiles = DragQueryFile(hDrop, 0xFFFFFFFF, NULL, 0);
	if(uNumFiles == 0) {
		GlobalUnlock(stg.hGlobal);
		ReleaseStgMedium (&stg);
		return E_INVALIDARG;
	}
	// IĂt@C()𓾂
	if(DragQueryFile(hDrop, 0, name, MAX_PATH) == 0) {
		GlobalUnlock(stg.hGlobal);
		ReleaseStgMedium(&stg);
		return E_INVALIDARG;
	}
	GlobalUnlock(stg.hGlobal);
	ReleaseStgMedium(&stg);

	WIN32_FIND_DATA fd;
	HANDLE hFile;

	folder_mode = folderModeExt;
	hFile = FindFirstFile(name, &fd);
	if(hFile != INVALID_HANDLE_VALUE) {
		if(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
			folder_mode = folderModeFolder;
		}
		FindClose(hFile);
	}
	// WXg猻݂̐ݒǂݏo
	GetRegistry();

	if(folder_mode) {
		file_path = name;
	} else {
		// IĂt@C̊gq𓾂
		file_ext = ExtractExt(name, regexp_mode);
		// tH_𓾂
		file_path = ExtractDirectory(name);
	}
	return NOERROR;
}

//
//	j[ACe̒ǉ
//
STDMETHODIMP CShellExt::QueryContextMenu(HMENU hMenu, UINT indexMenu, UINT idCmdFirst, UINT idCmdLast, UINT uFlags)
{
	UINT idCmd = idCmdFirst;

	GetRegistry();
	if((uFlags & CMF_DEFAULTONLY) == 0) {
		// j[}
		if(CheckWindows7()) {
			InsertMenu(hMenu, indexMenu++, MF_STRING | MF_BYPOSITION, idCmd++, GetString(regexp_mode ? IDS_MENU_REGEXP_7 : IDS_MENU_WILD_7).c_str());
		} else {
			InsertMenu(hMenu, indexMenu++, MF_STRING | MF_BYPOSITION, idCmd++, GetString(regexp_mode ? IDS_MENU_REGEXP : IDS_MENU_WILD).c_str());
		}
		return ResultFromScode(MAKE_SCODE(SEVERITY_SUCCESS, 0, (USHORT)(idCmd - idCmdFirst)));
	}
	return NOERROR;
}

//
//	t@CIԂɂ鏈
//	XP ȑOp
//
void CShellExt::SelectFile(HWND hWnd)
{
	int n;
	typedef struct {
	    char name[MAX_PATH];
	    LV_ITEM lvi;
	} INFO;
	INFO *p;
	HWND hFileMap;
	string file_name;

	file_list_size = 0;

	file_list.clear();
	hFileMap = static_cast<HWND>(CreateFileMapping((HANDLE)-1, NULL, PAGE_READWRITE, 0, 0x1000, "MyMap"));
	if(hFileMap){
		p = reinterpret_cast<INFO *>(MapViewOfFile(hFileMap, FILE_MAP_ALL_ACCESS, 0, 0, 0));
		n = -1;
		while((n = ListView_GetNextItem(hWnd, n, LVNI_ALL)) > -1) {
			p->lvi.mask = LVIF_TEXT;
			p->lvi.iItem = n;
			p->lvi.iSubItem = 0;
			p->lvi.cchTextMax = sizeof(p->name);
			p->lvi.pszText = p->name;
			ListView_GetItem(hWnd, reinterpret_cast<LPARAM>(&p->lvi));

			p->lvi.mask = LVIF_STATE;
			// ChJ[hňvH
			if(WildCard(wild_name, p->name, regexp_mode)) {
				p->lvi.state = LVIS_SELECTED;
				if(copy_flag >= modeCopyDirect && copy_flag <= modeCut) {
					file_name = string(file_path);
					file_name += "\\";
					file_name += p->name;
					file_list.push_back(file_name);
					file_list_size += file_name.size() + 1;
				}
			} else {
				p->lvi.state = 0;
			}
			p->lvi.stateMask = LVIS_SELECTED;
			ListView_SetItem(hWnd, (LPARAM)&p->lvi);
		}
		ListView_RedrawItems(hWnd, 0, ListView_GetItemCount(hWnd));

		UpdateWindow(hWnd);
		UnmapViewOfFile(p);
		CloseHandle(hFileMap);
	}
}

//
//	t@CIԂɂ鏈
//	XP/Vista p
//
void CShellExt::SelectFileXP(HWND hWnd)
{
	int n;
	char name[MAX_PATH];
	DWORD dwProcessId = 0;
	HANDLE hProcess;
    LVITEM lvItem;
	string file_name;

	file_list_size = 0;

	GetWindowThreadProcessId(hWnd, &dwProcessId);
	hProcess = OpenProcess(PROCESS_VM_OPERATION | PROCESS_VM_READ | PROCESS_VM_WRITE, FALSE, dwProcessId);

	file_list.clear();
	if(hProcess) {
		char *pvbuf = reinterpret_cast<char *>(VirtualAllocEx(hProcess, NULL, MAX_PATH + 1, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE));
		LVITEM *pvlvi = reinterpret_cast<LVITEM *>(VirtualAllocEx(hProcess, NULL, sizeof(LVITEM), MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE));

		n = -1;
		while((n = ListView_GetNextItem(hWnd, n, LVNI_ALL)) > -1) {
			lvItem.mask = LVIF_TEXT;
			lvItem.cchTextMax = MAX_PATH;
			lvItem.iItem = n;
			lvItem.iSubItem = 0;
			// vZX̃obt@ւ̃|C^w
			lvItem.pszText = pvbuf;
			// vZXɏ LVITEMցALVM_GETITEMp̃p^Rs[
			WriteProcessMemory(hProcess, pvlvi, &lvItem, sizeof(LVITEM), NULL);

			ListView_GetItem(hWnd, reinterpret_cast<LPARAM>(pvlvi));
			ReadProcessMemory(hProcess, pvbuf, name, MAX_PATH, NULL);

			lvItem.mask = LVIF_STATE;
			// ChJ[hňvH
			if(WildCard(wild_name, name, regexp_mode)) {
				lvItem.state = LVIS_SELECTED;
				lvItem.iItem = n;
				lvItem.iSubItem = 0;
				if(copy_flag >= modeCopyDirect && copy_flag <= modeCut) {
					file_name = string(file_path);
					file_name += "\\";
					file_name += name;
					file_list.push_back(file_name);
					file_list_size += file_name.size() + 1;
				}
			} else {
				lvItem.state = 0;
			}
			lvItem.stateMask = LVIS_SELECTED;
			lvItem.iItem = n;
			WriteProcessMemory(hProcess, pvlvi, &lvItem, sizeof(LVITEM), NULL);
			// ListView_SetItemState() gƁAX^bN LVITEM g
			// ܂̂ł܂ȂBłȂY񂾁B
			SendMessage(hWnd, LVM_SETITEMSTATE, static_cast<WPARAM>(n), reinterpret_cast<LPARAM>(pvlvi));
		}
		ListView_RedrawItems(hWnd, 0, ListView_GetItemCount(hWnd));

		VirtualFreeEx(hProcess, pvlvi, 0, MEM_RELEASE);
		VirtualFreeEx(hProcess, pvbuf, 0, MEM_RELEASE);

		CloseHandle(hProcess);
	}
}

//
//	WXg̒l𓾂
//
void CShellExt::GetRegistry(void)
{
	HKEY hKey;
	DWORD length, reg_type;
	int no;
	char name[MAX_PATH], buff[MAX_PATH];

	if(RegOpenKeyEx(HKEY_CURRENT_USER, CPFSHELL_REG, 0, KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS) {
		length = sizeof(DWORD);
		if(RegQueryValueEx(hKey, REG_EXP, NULL, &reg_type, reinterpret_cast<BYTE *>(&regexp_mode), &length) != ERROR_SUCCESS) {
			regexp_mode = FALSE;
		}
		length = sizeof(DWORD);
		if(RegQueryValueEx(hKey, RegModeList[folder_mode], NULL, &reg_type, reinterpret_cast<BYTE *>(&copy_mode[folder_mode]), &length) == ERROR_SUCCESS) {
			if(copy_mode[folder_mode] == modeLast) {
				length = sizeof(DWORD);
				RegQueryValueEx(hKey, RegFlagList[folder_mode], NULL, &reg_type, reinterpret_cast<BYTE *>(&copy_mode[folder_mode]), &length);
			}
			if(CheckWindows7() && copy_mode[folder_mode] == modeSelectOnly) {
				copy_mode[folder_mode] = modeCopyDirect;
			}
		} else {
			if(folder_mode == folderModeExt && !CheckWindows7()) {
				copy_mode[folder_mode] = modeSelectOnly;
			} else {
				copy_mode[folder_mode] = modeCopyDirect;
			}
		}
		length = sizeof(DWORD);
		if(RegQueryValueEx(hKey, REG_SELECT, NULL, &reg_type, reinterpret_cast<BYTE *>(&select_mode), &length) != ERROR_SUCCESS) {
			select_mode = selectExt;
		}
		history_list[folder_mode].clear();
		for(no = 0 ; no < HISTORY_SIZE ; no++) {
			wsprintf(name, "h%d_%d", folder_mode, no);
			length = MAX_PATH;
			if(RegQueryValueEx(hKey, name, NULL, &reg_type, reinterpret_cast<BYTE *>(buff), &length) == ERROR_SUCCESS) {
				if(buff[0] != '\0') {
					history_list[folder_mode].push_back(buff);
				}
			} else {
				break;
			}
		}
		folder_list[folder_mode].clear();
		for(no = 0 ; no < HISTORY_SIZE ; no++) {
			wsprintf(name, "f%d_%d", folder_mode, no);
			length = MAX_PATH;
			if(RegQueryValueEx(hKey, name, NULL, &reg_type, reinterpret_cast<BYTE *>(buff), &length) == ERROR_SUCCESS) {
				if(buff[0] != '\0') {
					folder_list[folder_mode].push_back(buff);
				}
			} else {
				break;
			}
		}
		length = sizeof(DWORD);
		if(RegQueryValueEx(hKey, RegUnderList[folder_mode], NULL, &reg_type, reinterpret_cast<BYTE *>(&under_mode[folder_mode]), &length) != ERROR_SUCCESS) {
			under_mode[folder_mode] = 0;
		}
		length = sizeof(DWORD);
		if(RegQueryValueEx(hKey, RegAddList[folder_mode], NULL, &reg_type, reinterpret_cast<BYTE *>(&add_mode[folder_mode]), &length) != ERROR_SUCCESS) {
			add_mode[folder_mode] = 0;
		}
		length = sizeof(DWORD);
		if(RegQueryValueEx(hKey, RegSameList[folder_mode], NULL, &reg_type, reinterpret_cast<BYTE *>(&same_mode[folder_mode]), &length) == ERROR_SUCCESS) {
			if(same_mode[folder_mode] == sameLast) {
				length = sizeof(DWORD);
				RegQueryValueEx(hKey, RegSameFlagList[folder_mode], NULL, &reg_type, reinterpret_cast<BYTE *>(&same_mode[folder_mode]), &length);
			}
		} else {
			same_mode[folder_mode] = sameQuery;
		}
		RegCloseKey(hKey);
	} else {
		regexp_mode = FALSE;
		if(folder_mode == folderModeExt && !CheckWindows7()) {
			copy_mode[folder_mode] = modeSelectOnly;
		} else {
			copy_mode[folder_mode] = modeCopyDirect;
		}
		select_mode = selectExt;
		under_mode[folder_mode] = 0;
		add_mode[folder_mode] = 0;
		same_mode[folder_mode] = sameQuery;
	}
	copy_dir.empty();
}

//
//	WXgɒlݒ
//
void CShellExt::SetRegistry(void)
{
	HKEY hKey;
	DWORD length;
	int no;
	char key_name[MAX_PATH];

	if(RegCreateKeyEx(HKEY_CURRENT_USER, CPFSHELL_REG, 0, "", REG_OPTION_NON_VOLATILE, KEY_SET_VALUE, NULL, &hKey, &length) == ERROR_SUCCESS) {
		RegSetValueEx(hKey, RegFlagList[folder_mode], 0, REG_DWORD, reinterpret_cast<BYTE *>(&copy_flag), sizeof(DWORD));
		RegSetValueEx(hKey, RegSameFlagList[folder_mode], 0, REG_DWORD, reinterpret_cast<BYTE *>(&same_type), sizeof(DWORD));
		no = 0;
		wsprintf(key_name, "h%d_%d", folder_mode, no++);
		RegSetValueEx(hKey, key_name, 0, REG_SZ, reinterpret_cast<BYTE *>(const_cast<char *>(wild_name.c_str())), wild_name.size() + 1);
		for(vector<string>::iterator name = history_list[folder_mode].begin() ; name != history_list[folder_mode].end() && no < 10 ; name++) {
			if(*name != wild_name && !wild_name.empty()) {
				wsprintf(key_name, "h%d_%d", folder_mode, no++);
				RegSetValueEx(hKey, key_name, 0, REG_SZ, reinterpret_cast<BYTE *>(const_cast<char *>(name->c_str())), name->size() + 1);
			}
		}
		if(copy_flag == modeCopyDirect || copy_flag == modeMove) {
			no = 0;
			wsprintf(key_name, "f%d_%d", folder_mode, no++);
			RegSetValueEx(hKey, key_name, 0, REG_SZ, reinterpret_cast<BYTE *>(const_cast<char *>(copy_dir_org.c_str())), copy_dir_org.size() + 1);
			for(vector<string>::iterator name = folder_list[folder_mode].begin() ; name != folder_list[folder_mode].end() && no < 10 ; name++) {
				if(*name != copy_dir_org && !copy_dir_org.empty()) {
					wsprintf(key_name, "f%d_%d", folder_mode, no++);
					RegSetValueEx(hKey, key_name, 0, REG_SZ, reinterpret_cast<BYTE *>(const_cast<char *>(name->c_str())), name->size() + 1);
				}
			}
		}
		RegCloseKey(hKey);
	}
}

//
//	ListView 
//
BOOL CALLBACK CShellExt::EnumChildProc(HWND hWnd, LPARAM lParam)
{
	char name[MAX_PATH];

	GetClassName(hWnd, name, MAX_PATH);
	if(!lstrcmp(name, "SysListView32")) {
		*reinterpret_cast<HWND *>(lParam) = hWnd;
		return FALSE;
	}
	return TRUE;
}

//
//	R}hs
//
STDMETHODIMP CShellExt::InvokeCommand(LPCMINVOKECOMMANDINFO lpcmi)
{
	UINT idCmd = LOWORD(lpcmi->lpVerb);
	HWND hWnd;

	if(!HIWORD(lpcmi->lpVerb)) {
		if(idCmd == 0) {
			hWnd = NULL;
			EnumChildWindows(lpcmi->hwnd, &EnumChildProc, reinterpret_cast<LPARAM>(&hWnd));
			if(hWnd != NULL || CheckWindows7()) {
				// WXgǂݏo
				GetRegistry();
				// _CAO\
				if(DialogBox(hInstanceDll, MAKEINTRESOURCE(IDD_DIALOG_CPF), lpcmi->hwnd, reinterpret_cast<DLGPROC>(InputFunc)) == IDOK) {
					CoInitialize(NULL);
					if(folder_mode) {
						AddFile(file_path);
						DirectCopy(hWnd);
					} else {
						if(CheckWindows7()) {
							AddFile(file_path);
						} else {
							// t@C̑I
							if(CheckWindowsXP()) {
								SelectFileXP(hWnd);
							} else {
								SelectFile(hWnd);
							}
						}
						// Nbv{[hփRs[邩H
						if(copy_flag == modeCopy || copy_flag == modeCut) {
							FileToClipboard(GetDesktopWindow());
						} else if(copy_flag == modeCopyDirect || copy_flag == modeMove) {
							DirectCopy(hWnd);
						}
					}
					SetFocus(hWnd);

					// WXgݒ
					SetRegistry();
					CoUninitialize();
				}
			} else {
				MessageBox(lpcmi->hwnd, GetString(IDS_ERROR_EXPLORER_WINDOW).c_str(), GetString(IDS_CAPTION).c_str(), MB_OK | MB_ICONERROR);
			}
		}
	}
	return NOERROR;
}

//
//	Xe[^X̕\
//
STDMETHODIMP CShellExt::GetCommandString(UINT idCmd, UINT uFlags, UINT FAR *reserved, LPSTR pszName, UINT cchMax)
{
	OSVERSIONINFO os;

	if(idCmd == 0) {
		os.dwOSVersionInfoSize = sizeof(os);
		GetVersionEx(&os);
		if(os.dwPlatformId == VER_PLATFORM_WIN32_NT) {
			MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, GetString(regexp_mode ? IDS_COMMAND_REGEXP : IDS_COMMAND_WILD).c_str(), -1, reinterpret_cast<LPWSTR>(pszName), cchMax);
		} else {
			lstrcpyn(pszName, GetString(regexp_mode ? IDS_COMMAND_REGEXP : IDS_COMMAND_WILD).c_str(), cchMax);
		}
	}
	return NOERROR;
}

//
//	t@CNbv{[hփRs[
//
void CShellExt::FileToClipboard(HWND hWnd)
{
	DROPFILES dfs;
	HGLOBAL h_name, h_flag;
	DWORD *flag_pt;
	char *pt;

	ZeroMemory(&dfs, sizeof(DROPFILES));
	dfs.pFiles = sizeof(dfs);
	h_name = GlobalAlloc(GHND, sizeof(DROPFILES) + file_list_size + 1);
	pt = reinterpret_cast<char *>(GlobalLock(h_name));
	memcpy(pt, &dfs, sizeof(DROPFILES));
	pt += sizeof(DROPFILES);
	for(vector<string>::iterator name = file_list.begin() ; name != file_list.end() ; name++) {
		lstrcpy(pt, name->c_str());
		pt += name->size() + 1;
	}
	GlobalUnlock(h_name);

	h_flag = GlobalAlloc(GHND, sizeof(DWORD));
	flag_pt = (DWORD *)GlobalLock(h_flag);
	*flag_pt = (copy_flag == modeCopy) ? DROPEFFECT_COPY: DROPEFFECT_MOVE;
	GlobalUnlock(h_flag);

	OpenClipboard(hWnd);
	EmptyClipboard();
	SetClipboardData(CF_HDROP, h_name);
	SetClipboardData(RegisterClipboardFormat(preferred_dropeffect), h_flag);
	CloseClipboard();
}

//
//	tH_I_CAÕR[obN
//
int CALLBACK BrowseCallbackProc(HWND hWnd, UINT uMsg, LPARAM lParam, LPARAM lpData)
{
	if(uMsg == BFFM_INITIALIZED) {
   		SendMessage(hWnd, BFFM_SETSELECTION, 1, lpData);
	}
	return 0;
}

//
//	Rs[̉
//
BOOL APIENTRY CopyFunc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message) {
	case WM_INITDIALOG:
		{
			MouseCursorWindow(hDlg, 0);

			HWND hCtrl = GetDlgItem(hDlg, IDC_ANIMATE_COPY);
			// Vista ȍ~ł shell32.dll ̃\[X avi ݂Ȃߓ̂gp
			if(CheckWindowsVista()) {
				if(pShellExt->GetCopyFlag() == modeCopyDirect) {
					SendMessage(hCtrl, ACM_OPEN, reinterpret_cast<WPARAM>(hInstanceDll), static_cast<LPARAM>(IDR_COPY));
				} else {
					SendMessage(hCtrl, ACM_OPEN, reinterpret_cast<WPARAM>(hInstanceDll), static_cast<LPARAM>(IDR_MOVE));
				}
			} else {
				HINSTANCE hAvi = LoadLibrary("shell32.dll");
				if(pShellExt->GetCopyFlag() == modeCopyDirect) {
					SendMessage(hCtrl, ACM_OPEN, reinterpret_cast<WPARAM>(hAvi), 161);
				} else {
					SendMessage(hCtrl, ACM_OPEN, reinterpret_cast<WPARAM>(hAvi), 160);
				}
			}
			Animate_Play(hCtrl, 0, -1, -1);
			hCtrl = GetDlgItem(hDlg, IDC_PROGRESS_COPY);
			SendMessage(hCtrl, PBM_SETRANGE32, 0, pShellExt->GetCopyCount());
		}
		break;

	case WM_PROGRESS_COPY:
		{
			HWND hCtrl = GetDlgItem(hDlg, IDC_PROGRESS_COPY);
			SendMessage(hCtrl, PBM_SETPOS, wParam, 0);
			string name = pShellExt->GetDestName();
			string str = ExtractFilename(name);
			str += GetString(IDS_PROGRESS_TO);
			str += ExtractDirectory(name, true);
			str += GetString(pShellExt->GetCopyFlag() == modeCopyDirect ? IDS_PROGRESS_COPY : IDS_PROGRESS_MOVE);
			hCtrl = GetDlgItem(hDlg, IDC_STATIC_COPY);
			SetWindowText(hCtrl, str.c_str());
		}
		break;

	case WM_COMMAND:
		switch(LOWORD (wParam)) {
		case IDCANCEL:
			pShellExt->SetCancelFlag();
			break;
		}
		break;
	}
	return FALSE;
}

//
//	㏑₢킹
//
BOOL APIENTRY QueryFunc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message) {
	case WM_INITDIALOG:
		{
			CenterParentWindow(hDlg);

			char *text = reinterpret_cast<char *>(lParam);
			SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT), text);
		}
		break;
	case WM_COMMAND:
		switch(LOWORD (wParam)) {
		case IDOK:
			EndDialog(hDlg, IDOK);
			break;
		case IDCANCEL:
			EndDialog(hDlg, IDCANCEL);
			break;
		case IDC_BUTTON_ALL:
			EndDialog(hDlg, IDIGNORE);
			break;
		}
	}
	return FALSE;
}

//
//	wtH_̈vt@CXgɒǉ
//
void CShellExt::AddFile(string path)
{
	WIN32_FIND_DATA fd;
	HANDLE hFile;
	string name;

	file_list_size = 0;
	name = path;
	name += "\\*.*";
	hFile = FindFirstFile(name.c_str(), &fd);
	if(hFile != INVALID_HANDLE_VALUE) {
		do {
			if((fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 && WildCard(wild_name, fd.cFileName, regexp_mode)) {
				name = path;
				name += "\\";
				name += fd.cFileName;
				file_list.push_back(name);
				file_list_size += name.size() + 1;
			}
		} while(FindNextFile(hFile, &fd));
		FindClose(hFile);
	}
}

//
//	wtH_̃TutH_̈vt@CXgɒǉ
//
void CShellExt::AddSubFile(string path, BOOL flag)
{
	WIN32_FIND_DATA fd;
	HANDLE hFile;
	string name;

	name = path;
	name += "\\*.*";
	hFile = FindFirstFile(name.c_str(), &fd);
	if(hFile != INVALID_HANDLE_VALUE) {
		do {
			if(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
				if(lstrcmp(fd.cFileName, ".") && lstrcmp(fd.cFileName, "..")) {
					name = path;
					name += "\\";
					name += fd.cFileName;
					AddSubFile(name, TRUE);
				}
			} else if(WildCard(wild_name, fd.cFileName, regexp_mode) && flag) {
				name = path;
				name += "\\";
				name += fd.cFileName;
				file_list.push_back(name);
			}
		} while(FindNextFile(hFile, &fd));
		FindClose(hFile);
	}
}

//
//	ڃRs[
//
void CShellExt::DirectCopy(HWND hWnd)
{
	string dir;
	int count;
	MSG msg;
	HWND hDlg;
	int all_flag = FALSE;

	if(under_type) {
		// TutH_̃t@CXgɒǉ
		AddSubFile(file_path, FALSE);
	}

	// i_CAO𐶐
	cancel_flag = FALSE;
	hDlg = CreateDialog(hInstanceDll, MAKEINTRESOURCE(IDD_DIALOG_COPY), hWnd, static_cast<DLGPROC>(CopyFunc));
	ShowWindow(hDlg, SW_SHOW);

	dir = DeleteYen(copy_dir);

	result_list.clear();
	count = 1;
	for(vector<string>::iterator name = file_list.begin() ; name != file_list.end() ; name++) {
		if(cancel_flag) {
			break;
		}
		// Rs[̃t@C
		dst_name = dir + name->substr(file_path.size());
		// vOXo[i߂
		PostMessage(hDlg, WM_PROGRESS_COPY, count++, 0);
		while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			if(!IsDialogMessage(hDlg, &msg)) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		WIN32_FIND_DATA src, dst;
		HANDLE hFile;
		BOOL flag = TRUE;
		// ㏑̖₢킹
		if((same_type == sameQuery || same_type == sameNew) && !all_flag) {
			flag = FALSE;
			hFile = FindFirstFile(name->c_str(), &src);
			if(hFile != INVALID_HANDLE_VALUE) {
				FindClose(hFile);
				hFile = FindFirstFile(dst_name.c_str(), &dst);
				if(hFile == INVALID_HANDLE_VALUE) {
					flag = TRUE;
				} else {
					FindClose(hFile);
					if(same_type == sameNew) {
						// Vǂ
						if(src.ftLastWriteTime.dwHighDateTime > dst.ftLastWriteTime.dwHighDateTime) {
							flag = TRUE;
						} else if(src.ftLastWriteTime.dwHighDateTime == dst.ftLastWriteTime.dwHighDateTime) {
							if(src.ftLastWriteTime.dwLowDateTime > dst.ftLastWriteTime.dwLowDateTime) {
								flag = TRUE;
							}
						}
					} else {
						FILETIME fsrc, fdst;
						SYSTEMTIME ssrc, sdst;
						ostringstream os;

						FileTimeToLocalFileTime(&src.ftLastWriteTime, &fsrc);
						FileTimeToLocalFileTime(&src.ftLastWriteTime, &fdst);
						FileTimeToSystemTime(&fsrc, &ssrc);
						FileTimeToSystemTime(&fdst, &sdst);

						// ₢킹p쐬
						os << GetString(IDS_SAME_MESSAGE) << "\n\n";
						os << GetString(IDS_SAME_SRC_FILE) << ExtractFilename(*name) << "\n";
						os << GetString(IDS_SAME_SRC_DIR) << ExtractDirectory(*name, true) << "\n";
						os << GetString(IDS_SAME_SRC_DATE) << setw(2) << setfill('0') << ssrc.wYear << "/" << setw(2) << setfill('0') << ssrc.wMonth << "/" << setw(2) << setfill('0') << ssrc.wDay << " ";
						os << setw(2) << setfill('0') << ssrc.wHour << ":" << setw(2) << setfill('0') << ssrc.wMinute << ":" << setw(2) << setfill('0') << ssrc.wSecond << "\n";
						os << GetString(IDS_SAME_SRC_SIZE) << src.nFileSizeLow << GetString(IDS_SAME_BYTES);
						os << "\n\n";
						os << GetString(IDS_SAME_SRC_TO_DST);
						os << "\n\n";
						os << GetString(IDS_SAME_DST_FILE) << ExtractFilename(dst_name) << "\n";
						os << GetString(IDS_SAME_DST_DIR) << ExtractDirectory(dst_name, true) << "\n";
						os << GetString(IDS_SAME_DST_DATE) << setw(2) << setfill('0') << sdst.wYear << "/" << setw(2) << setfill('0') << sdst.wMonth << "/" << setw(2) << setfill('0') << sdst.wDay << " ";
						os << setw(2) << setfill('0') << sdst.wHour << ":" << setw(2) << setfill('0') << setw(2) << setfill('0') << sdst.wMinute << ":" << sdst.wSecond << "\n";
						os << GetString(IDS_SAME_DST_SIZE) << dst.nFileSizeLow;
						os << GetString(IDS_SAME_BYTES);
						os << "\n\n";
						os << GetString(IDS_SAME_QUERY);
						int res = DialogBoxParam(hInstanceDll, MAKEINTRESOURCE(IDD_DIALOG_QUERY), hDlg, reinterpret_cast<DLGPROC>(QueryFunc), (long)os.str().c_str());
						if(res == IDOK || res == IDIGNORE) {
							flag = TRUE;
							if(res == IDIGNORE) {
								all_flag = TRUE;
							}
						}
					}
				}
			}
		}
		if(flag) {
			COPY_RESULT result;

			result.flag = FALSE;
			hFile = FindFirstFile(ExtractDirectory(dst_name).c_str(), &dst);
			if(hFile != INVALID_HANDLE_VALUE) {
				FindClose(hFile);
			} else {
				// tH_쐬
				ForceDirectories(ExtractDirectory(dst_name));
			}
			result.src = ExtractDirectory(*name, true);
			result.dst = ExtractDirectory(dst_name, true);
			result.file = ExtractFilename(*name);
			if(copy_flag == modeMove) {
				// ړ̏ꍇ
				if(MoveFileEx(name->c_str(), dst_name.c_str(), MOVEFILE_REPLACE_EXISTING | MOVEFILE_COPY_ALLOWED)) {
					result.flag = TRUE;
				}
			} else if(CopyFile(name->c_str(), dst_name.c_str(), FALSE)) {
				result.flag = TRUE;
			}
			result_list.push_back(result);
		}
	}
	DestroyWindow(hDlg);

	// ʃ_CAO\
	DialogBox(hInstanceDll, MAKEINTRESOURCE(IDD_DIALOG_LOG), hWnd, reinterpret_cast<DLGPROC>(LogFunc));

	result_list.clear();
}

//
//	Ԃ
//
string CShellExt::GetFileExt(void)
{
	if(select_mode == selectExt && !folder_mode) {
		return file_ext;
	} else if(!history_list[folder_mode].empty()) {
		return *history_list[folder_mode].begin();
	}
	return "";
}

//
//	R{{bNXɗǉ
//
void CShellExt::SetComboBox(HWND hCtrl)
{
	for(vector<string>::iterator name = history_list[folder_mode].begin() ; name != history_list[folder_mode].end() ; name++) {
		ComboBox_AddString(hCtrl, name->c_str());
	}
}

//
//	R{{bNXɃtH_ǉ
//
void CShellExt::SetFolderComboBox(HWND hCtrl)
{
	for(vector<string>::iterator name = folder_list[folder_mode].begin() ; name != folder_list[folder_mode].end() ; name++) {
		ComboBox_AddString(hCtrl, name->c_str());
	}
}

//
//	݂̑I[h擾
//
int GetSelectMode(HWND hDlg)
{
	int flag = modeSelectOnly;
	if(Button_GetCheck(GetDlgItem(hDlg, IDC_RADIO_CUT))) {
		flag = modeCut;
	} else if(Button_GetCheck(GetDlgItem(hDlg, IDC_RADIO_COPY))) {
		flag = modeCopy;
	} else if(Button_GetCheck(GetDlgItem(hDlg, IDC_RADIO_COPY_DIRECT))) {
		flag = modeCopyDirect;
	} else if(Button_GetCheck(GetDlgItem(hDlg, IDC_RADIO_MOVE))) {
		flag = modeMove;
	}
	return flag;
}

//
//	̍쐬
//
void SetCommentText(HWND hDlg, WPARAM wParam)
{
	HWND hCtrl;
	vector<char> text;
	ostringstream os;
	int mode, no;
	int under_flag = IsWindowEnabled(GetDlgItem(hDlg, IDC_CHECK_UNDER)) && Button_GetCheck(GetDlgItem(hDlg, IDC_CHECK_UNDER));
	int add_flag = IsWindowEnabled(GetDlgItem(hDlg, IDC_CHECK_ADD)) && Button_GetCheck(GetDlgItem(hDlg, IDC_CHECK_ADD));

	mode = GetSelectMode(hDlg);
	os << GetString(IDS_COMMENT_FOLDER) << pShellExt->GetFilePath() << " " << ((under_flag) ? GetString(IDS_COMMENT_SUBFOLDER) : "") << GetString(IDS_COMMENT_NAME);
	os << (pShellExt->GetRegexpMode() ? GetString(IDS_REGEXP) : GetString(IDS_WILD));
	hCtrl = GetDlgItem(hDlg, IDC_COMBO_NAME);
	no = ComboBox_GetCurSel(hCtrl);
	if(LOWORD(wParam) == IDC_COMBO_NAME && HIWORD(wParam) == CBN_SELCHANGE && no != CB_ERR) {
		text.reserve(ComboBox_GetLBTextLen(hCtrl, no) + 1);
		ComboBox_GetLBText(hCtrl, no, &text[0]);
	} else {
		text.reserve(MAX_PATH + 1);
		GetWindowText(hCtrl, &text[0], MAX_PATH);
	}
	os << " " << reinterpret_cast<char *>(&text[0]) << GetString(IDS_COMMENT_MATCH);
	if(mode == modeCopyDirect || mode == modeMove) {
		hCtrl = GetDlgItem(hDlg, IDC_COMBO_FOLDER);
		no = ComboBox_GetCurSel(hCtrl);
		if(LOWORD(wParam) == IDC_COMBO_FOLDER && HIWORD(wParam) == CBN_SELCHANGE && no != CB_ERR) {
			text.reserve(ComboBox_GetLBTextLen(hCtrl, no) + 1);
			ComboBox_GetLBText(hCtrl, no, &text[0]);
		} else {
			text.reserve(MAX_PATH + 1);
			GetWindowText(hCtrl, &text[0], MAX_PATH);
		}
		int yen_flag = TRUE;
		if(add_flag) {
			string str = ExtractFilename(pShellExt->GetFilePath());
			if(!str.empty()) {
				os << " " << DeleteYen(&text[0]);
				os << "\\" << str;
				yen_flag = FALSE;
			}
		}
		if(yen_flag) {
			os << " " << reinterpret_cast<char *>(&text[0]);
		}
		os << " " << GetString((mode == modeCopyDirect) ? IDS_COPY : IDS_MOVE) << GetString(IDS_COMMENT_DO);
		if(pShellExt->GetFolderMode()) {
			os << "\n" << GetString(IDS_COMMENT_NO_SELECT);
		} else if(under_flag) {
			os << "\n" << GetString(IDS_COMMENT_NO_SELECT_SUBFOLDER);
		}
		os << "\n" << GetString((mode == modeCopyDirect) ? IDS_COPY : IDS_MOVE) << GetString(IDS_COMMENT_MAKE_FOLDER);
	} else if(mode == modeCopy || mode == modeCut) {
		os << GetString(IDS_COMMENT_CLIPBOARD) << GetString((mode == modeCopy) ? IDS_COMMENT_CLIPBOARD_COPY : IDS_COMMENT_CLIPBOARD_CUT) << GetString(IDS_COMMENT_CLIPBOARD_DO);
	} else {
		os << GetString(IDS_COMMENT_SELECT);
	}
	SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT), os.str().c_str());
}

//
//	݂̐ݒɍ킹ĉʂ̏ԂύX
//
void SetInputItem(HWND hDlg)
{
	HWND hCtrl;
	int mode;
	string text;

	if(pShellExt->GetRegexpMode()) {
		text = GetString(IDS_CAPTION_REGEXP);
	} else {
		text = GetString(IDS_CAPTION_WILDCARD);
	}
	if(pShellExt->GetFolderMode()) {
		text += GetString(IDS_CAPTION_FOLDER_MODE);
	}
	SetWindowText(hDlg, text.c_str());

	hCtrl = GetDlgItem(hDlg, IDC_STATIC_WILDCARD);
	if(pShellExt->GetRegexpMode()) {
		SetWindowText(hCtrl, GetString(IDS_STATIC_REGEXP).c_str());
	} else {
		SetWindowText(hCtrl, GetString(IDS_STATIC_WILDCARD).c_str());
	}

	// Nbv{[hRs[[h̏Ԃɂ킹
	// WI{^Zbg
	mode = pShellExt->GetCopyMode();
	if(pShellExt->GetFolderMode() && mode != modeCopyDirect && mode != modeMove) {
		mode = modeCopyDirect;
	}
	if(CheckWindows7()) {
		if(mode == modeSelectOnly) {
			mode = modeCopyDirect;
		}
	}
	hCtrl = GetDlgItem(hDlg, IDC_RADIO_COPY_DIRECT);
	Button_SetCheck(hCtrl, (mode == modeCopyDirect) ? BST_CHECKED : BST_UNCHECKED);
	hCtrl = GetDlgItem(hDlg, IDC_RADIO_MOVE);
	Button_SetCheck(hCtrl, (mode == modeMove) ? BST_CHECKED : BST_UNCHECKED);
	hCtrl = GetDlgItem(hDlg, IDC_RADIO_COPY);
	Button_SetCheck(hCtrl, (mode == modeCopy) ? BST_CHECKED : BST_UNCHECKED);
	EnableWindow(hCtrl, !pShellExt->GetFolderMode());
	hCtrl = GetDlgItem(hDlg, IDC_RADIO_CUT);
	Button_SetCheck(hCtrl, (mode == modeCut) ? BST_CHECKED : BST_UNCHECKED);
	EnableWindow(hCtrl, !pShellExt->GetFolderMode());
	hCtrl = GetDlgItem(hDlg, IDC_RADIO_NONE);
	Button_SetCheck(hCtrl, (mode == modeSelectOnly) ? BST_CHECKED : BST_UNCHECKED);
	EnableWindow(hCtrl, !pShellExt->GetFolderMode() && !CheckWindows7());

	// tH_R{{bNXɗZbg
	hCtrl = GetDlgItem(hDlg, IDC_COMBO_FOLDER);
	ComboBox_ResetContent(hCtrl);
	EnableWindow(hCtrl, (mode == modeCopyDirect || mode == modeMove));
	pShellExt->SetFolderComboBox(hCtrl);
	ComboBox_SetCurSel(hCtrl, 0);

	EnableWindow(GetDlgItem(hDlg, IDC_BUTTON_FOLDER), (mode == modeCopyDirect || mode == modeMove));
	EnableWindow(GetDlgItem(hDlg, IDC_CHECK_UNDER), (mode == modeCopyDirect || mode == modeMove));
	EnableWindow(GetDlgItem(hDlg, IDC_STATIC_SAME), (mode == modeCopyDirect || mode == modeMove));
	EnableWindow(GetDlgItem(hDlg, IDC_COMBO_SAME), (mode == modeCopyDirect || mode == modeMove));
	EnableWindow(GetDlgItem(hDlg, IDC_CHECK_ADD), (mode == modeCopyDirect || mode == modeMove));

	// Zbg
	hCtrl = GetDlgItem(hDlg, IDC_COMBO_NAME);
	ComboBox_ResetContent(hCtrl);
	std::string name = pShellExt->GetFileExt();
	SetWindowText(hCtrl, name.c_str());
	// gqR{{bNXɗZbg
	pShellExt->SetComboBox(hCtrl);

	SetFocus(hCtrl);

	SetCommentText(hDlg, 0);
}

//
//	HTML wv\
//
void ShowHtmlHelp(HWND hWnd, int no, int id)
{
	char name[MAX_PATH];
	string path;

	GetModuleFileName(hInstanceDll, name, MAX_PATH);
	path = ExtractDirectory(name);
	path += "\\cpfShell.chm";
	HtmlHelp(hWnd, path.c_str(), no, id);
}

//
//	_CAObZ[W
//
BOOL APIENTRY InputFunc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message) {
	case WM_INITDIALOG:
		{
			HICON hIcon;
			if(hIcon = LoadIcon(hInstanceDll, MAKEINTRESOURCE(IDI_ICON_CPF))) {
				SendMessage(hDlg, WM_SETICON, ICON_SMALL, (LPARAM)hIcon);
				SendMessage(hDlg, WM_SETICON, ICON_BIG, (LPARAM)hIcon);
			}
			MouseCursorWindow(hDlg, 0);
			// ݒ
			string text = " " + pShellExt->GetFilePath();
			SetWindowText(GetDlgItem(hDlg, IDC_STATIC_FOLDER), text.c_str());

			Button_SetCheck(GetDlgItem(hDlg, IDC_CHECK_UNDER), pShellExt->GetUnderMode());
			Button_SetCheck(GetDlgItem(hDlg, IDC_CHECK_ADD), pShellExt->GetAddMode());

			SetInputItem(hDlg);

			HWND hCtrl = GetDlgItem(hDlg, IDC_COMBO_SAME);
			ComboBox_ResetContent(hCtrl);
			ComboBox_AddString(hCtrl, GetString(IDS_OVERWRITE_QUERY).c_str());
			ComboBox_AddString(hCtrl, GetString(IDS_OVERWRITE_NEW).c_str());
			ComboBox_AddString(hCtrl, GetString(IDS_OVERWRITE_NO_QUERY).c_str());
			ComboBox_SetCurSel(hCtrl, pShellExt->GetSameMode());
		}
		break;

	case WM_COMMAND:
		switch(LOWORD(wParam)) {
		case IDOK:
			{
				char name[MAX_PATH], temp[MAX_PATH];
				string dir;
				int mode;
				// Rs[tOݒ
				mode = GetSelectMode(hDlg);
				// ChJ[hݒ
				GetWindowText(GetDlgItem(hDlg, IDC_COMBO_NAME), name, MAX_PATH);
				if(name[0] == '\0') {
					MessageBox(hDlg, GetString(pShellExt->GetRegexpMode() ? IDS_ERROR_NO_REGEXP_STRING : IDS_ERROR_NO_WILDCARD_STRING).c_str(), GetString(IDS_CAPTION).c_str(), MB_OK | MB_ICONERROR);
					break;
				}
				// tH_ݒ
				GetWindowText(GetDlgItem(hDlg, IDC_COMBO_FOLDER), temp, MAX_PATH);
				if((mode == modeCopyDirect || mode == modeMove) && temp[0] == '\0') {
					MessageBox(hDlg, GetString((mode == modeCopyDirect) ? IDS_ERROR_NO_COPY_FOLDER : IDS_ERROR_NO_MOVE_FOLDER).c_str(), GetString(IDS_CAPTION).c_str(), MB_OK | MB_ICONERROR);
					break;
				}
				pShellExt->SetCopyFlag(mode);
				pShellExt->SetWildName(name);
				dir = temp;
				pShellExt->SetCopyDirOrg(dir);
				if(Button_GetCheck(GetDlgItem(hDlg, IDC_CHECK_ADD))) {
					dir = DeleteYen(dir) + "\\" + ExtractFilename(pShellExt->GetFilePath());
				}
				pShellExt->SetCopyDir(dir);
				pShellExt->SetUnderType(Button_GetCheck(GetDlgItem(hDlg, IDC_CHECK_UNDER)));
				pShellExt->SetSameType(ComboBox_GetCurSel(GetDlgItem(hDlg, IDC_COMBO_SAME)));

				EndDialog(hDlg, IDOK);
			}
			break;
		case IDCANCEL:
			EndDialog(hDlg, IDCANCEL);
			break;
		case IDC_BUTTON_FOLDER:
			{
				BROWSEINFO bri;
				LPITEMIDLIST lpid;
				string str, dir;

				str = GetString(Button_GetCheck(GetDlgItem(hDlg, IDC_RADIO_COPY_DIRECT)) ? IDS_SELECT_COPY_FOLDER : IDS_SELECT_MOVE_FOLDER);
				dir = pShellExt->GetFilePath();
				ZeroMemory(&bri, sizeof(BROWSEINFO));
				bri.hwndOwner = hDlg;
				bri.lpszTitle = const_cast<char *>(str.c_str());
				bri.pidlRoot = NULL;
				bri.lpfn = BrowseCallbackProc;
				bri.ulFlags = BIF_RETURNONLYFSDIRS;
				bri.lParam = reinterpret_cast<LPARAM>(dir.c_str());
				if(lpid = SHBrowseForFolder(&bri)) {
					char name[MAX_PATH];
					SHGetPathFromIDList(lpid, name);
					CoTaskMemFree(lpid);
					SetWindowText(GetDlgItem(hDlg, IDC_COMBO_FOLDER), name);
				}
			}
			break;
		case IDC_BUTTON_OPTION:
			if(DialogBox(hInstanceDll, MAKEINTRESOURCE(IDD_DIALOG_OPTION), hDlg, reinterpret_cast<DLGPROC>(OptionFunc)) == IDOK) {
				pShellExt->GetRegistry();
				SetInputItem(hDlg);
			}
			break;
		case IDC_RADIO_COPY_DIRECT:
		case IDC_RADIO_MOVE:
		case IDC_RADIO_COPY:
		case IDC_RADIO_CUT:
		case IDC_RADIO_NONE:
			EnableWindow(GetDlgItem(hDlg, IDC_COMBO_FOLDER), LOWORD(wParam) >= IDC_RADIO_COPY_DIRECT);
			EnableWindow(GetDlgItem(hDlg, IDC_BUTTON_FOLDER), LOWORD(wParam) >= IDC_RADIO_COPY_DIRECT);
			EnableWindow(GetDlgItem(hDlg, IDC_CHECK_UNDER), LOWORD(wParam) >= IDC_RADIO_COPY_DIRECT);
			EnableWindow(GetDlgItem(hDlg, IDC_STATIC_SAME), LOWORD(wParam) >= IDC_RADIO_COPY_DIRECT);
			EnableWindow(GetDlgItem(hDlg, IDC_COMBO_SAME), LOWORD(wParam) >= IDC_RADIO_COPY_DIRECT);
			EnableWindow(GetDlgItem(hDlg, IDC_CHECK_ADD), LOWORD(wParam) >= IDC_RADIO_COPY_DIRECT);
			break;
		}
		SetCommentText(hDlg, wParam);
		break;
	case WM_HELP:
		ShowHtmlHelp(hDlg, HH_HELP_CONTEXT, HLP_IDH_MAIN);
		return TRUE;
	}
	return FALSE;
}

//
//	ݒ֘ÃWXg̒l𓾂
//
void get_registory(int &regexp_mode, int option[])
{
	HKEY hKey;
	DWORD length, reg_type;

	option[optionExtMode] = modeSelectOnly;
	option[optionExtSelect] = selectExt;
	option[optionExtUnder] = 0;
	option[optionExtAdd] = 0;
	option[optionExtSame] = sameQuery;
	option[optionFolderMode] = modeCopyDirect;
	option[optionFolderUnder] = 0;
	option[optionFolderAdd] = 0;
	option[optionFolderSame] = sameQuery;
	regexp_mode = FALSE;
	if(RegOpenKeyEx(HKEY_CURRENT_USER, CPFSHELL_REG, 0, KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS) {
		length = sizeof(DWORD);
		RegQueryValueEx(hKey, REG_EXP, NULL, &reg_type, reinterpret_cast<BYTE *>(&regexp_mode), &length);

		length = sizeof(DWORD);
		RegQueryValueEx(hKey, RegModeList[folderModeExt], NULL, &reg_type, reinterpret_cast<BYTE *>(&option[optionExtMode]), &length);
		length = sizeof(DWORD);
		RegQueryValueEx(hKey, REG_SELECT, NULL, &reg_type, reinterpret_cast<BYTE *>(&option[optionExtSelect]), &length);
		length = sizeof(DWORD);
		RegQueryValueEx(hKey, RegUnderList[folderModeExt], NULL, &reg_type, reinterpret_cast<BYTE *>(&option[optionExtUnder]), &length);
		length = sizeof(DWORD);
		RegQueryValueEx(hKey, RegAddList[folderModeExt], NULL, &reg_type, reinterpret_cast<BYTE *>(&option[optionExtAdd]), &length);
		length = sizeof(DWORD);
		RegQueryValueEx(hKey, RegSameList[folderModeExt], NULL, &reg_type, reinterpret_cast<BYTE *>(&option[optionExtSame]), &length);

		length = sizeof(DWORD);
		RegQueryValueEx(hKey, RegModeList[folderModeFolder], NULL, &reg_type, reinterpret_cast<BYTE *>(&option[optionFolderMode]), &length);
		length = sizeof(DWORD);
		RegQueryValueEx(hKey, RegUnderList[folderModeFolder], NULL, &reg_type, reinterpret_cast<BYTE *>(&option[optionFolderUnder]), &length);
		length = sizeof(DWORD);
		RegQueryValueEx(hKey, RegAddList[folderModeFolder], NULL, &reg_type, reinterpret_cast<BYTE *>(&option[optionFolderAdd]), &length);
		length = sizeof(DWORD);
		RegQueryValueEx(hKey, RegSameList[folderModeFolder], NULL, &reg_type, reinterpret_cast<BYTE *>(&option[optionFolderSame]), &length);

		RegCloseKey(hKey);
	}
}

//
//	ݒ֘A̒lWXgɏ
//
void set_registory(int regexp_mode, int option[])
{
	HKEY hKey;
	DWORD length;

	if(RegCreateKeyEx(HKEY_CURRENT_USER, CPFSHELL_REG, 0, "" , REG_OPTION_NON_VOLATILE, KEY_SET_VALUE, NULL, &hKey, &length) == ERROR_SUCCESS) {
		RegSetValueEx(hKey, REG_EXP, 0, REG_DWORD, reinterpret_cast<BYTE *>(&regexp_mode), sizeof(DWORD));

		RegSetValueEx(hKey, RegModeList[folderModeExt], 0, REG_DWORD, reinterpret_cast<BYTE *>(&option[optionExtMode]), sizeof(DWORD));
		RegSetValueEx(hKey, REG_SELECT, 0, REG_DWORD, reinterpret_cast<BYTE *>(&option[optionExtSelect]), sizeof(DWORD));
		RegSetValueEx(hKey, RegUnderList[folderModeExt], 0, REG_DWORD, reinterpret_cast<BYTE *>(&option[optionExtUnder]), sizeof(DWORD));
		RegSetValueEx(hKey, RegAddList[folderModeExt], 0, REG_DWORD, reinterpret_cast<BYTE *>(&option[optionExtAdd]), sizeof(DWORD));
		RegSetValueEx(hKey, RegSameList[folderModeExt], 0, REG_DWORD, reinterpret_cast<BYTE *>(&option[optionExtSame]), sizeof(DWORD));

		RegSetValueEx(hKey, RegModeList[folderModeFolder], 0, REG_DWORD, reinterpret_cast<BYTE *>(&option[optionFolderMode]), sizeof(DWORD));
		RegSetValueEx(hKey, RegUnderList[folderModeFolder], 0, REG_DWORD, reinterpret_cast<BYTE *>(&option[optionFolderUnder]), sizeof(DWORD));
		RegSetValueEx(hKey, RegAddList[folderModeFolder], 0, REG_DWORD, reinterpret_cast<BYTE *>(&option[optionFolderAdd]), sizeof(DWORD));
		RegSetValueEx(hKey, RegSameList[folderModeFolder], 0, REG_DWORD, reinterpret_cast<BYTE *>(&option[optionFolderSame]), sizeof(DWORD));
		RegCloseKey(hKey);
	}
}

//
//	_CAObZ[W
//
BOOL APIENTRY OptionFunc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message) {
	case WM_INITDIALOG:
		{
			int regexp_mode, option[optionMax];
			HICON hIcon;
			HWND hCtrl;
			int no;

			CenterParentWindow(hDlg);
			if(hIcon = LoadIcon(hInstanceDll, MAKEINTRESOURCE(IDI_ICON_CPF))) {
				SendMessage(hDlg, WM_SETICON, ICON_SMALL, reinterpret_cast<LPARAM>(hIcon));
				SendMessage(hDlg, WM_SETICON, ICON_BIG, reinterpret_cast<LPARAM>(hIcon));
			}
			// ݒlǂ
			get_registory(regexp_mode, option);

			// K\̐ݒ
			Button_SetCheck(GetDlgItem(hDlg, IDC_CHECK_REGEXP), regexp_mode);

			// Nbv{[hRs[[hݒ
			hCtrl = GetDlgItem(hDlg, IDC_COMBO_EXT_MODE);
			for(no = modeCopyDirect ; no <= modeLast ; no++) {
				if(!CheckWindows7() || no != modeSelectOnly) {
					ComboBox_AddString(hCtrl, GetString(IDS_MODE_DIRECT_COPY + no).c_str());
				}
			}
			if(CheckWindows7() && option[optionExtMode] == modeLast) {
				option[optionExtMode] = modeLast7;
			}
			ComboBox_SetCurSel(hCtrl, option[optionExtMode]);

			hCtrl = GetDlgItem(hDlg, IDC_COMBO_FOLDER_MODE);
			ComboBox_AddString(hCtrl, GetString(IDS_MODE_DIRECT_COPY).c_str());
			ComboBox_AddString(hCtrl, GetString(IDS_MODE_MOVE).c_str());
			ComboBox_AddString(hCtrl, GetString(IDS_MODE_LAST).c_str());
			if(option[optionFolderMode] <= modeMove) {
				ComboBox_SetCurSel(hCtrl, option[optionFolderMode]);
			} else {
				ComboBox_SetCurSel(hCtrl, 2);
			}

			// I𕶎񃂁[hWI{^ɐݒ
			hCtrl = GetDlgItem(hDlg, IDC_COMBO_SELECT);
			ComboBox_AddString(hCtrl, GetString(IDS_SELECT_EXT).c_str());
			ComboBox_AddString(hCtrl, GetString(IDS_SELECT_TOP).c_str());
			ComboBox_SetCurSel(hCtrl, option[optionExtSelect]);

			Button_SetCheck(GetDlgItem(hDlg, IDC_CHECK_EXT_UNDER), option[optionExtUnder]);
			Button_SetCheck(GetDlgItem(hDlg, IDC_CHECK_EXT_ADD), option[optionExtAdd]);
			Button_SetCheck(GetDlgItem(hDlg, IDC_CHECK_FOLDER_UNDER), option[optionFolderUnder]);
			Button_SetCheck(GetDlgItem(hDlg, IDC_CHECK_FOLDER_ADD), option[optionFolderAdd]);
			hCtrl = GetDlgItem(hDlg, IDC_COMBO_EXT_SAME);
			for(no = sameQuery ; no <= sameLast ; no++) {
				ComboBox_AddString(hCtrl, GetString(IDS_OVERWRITE_QUERY + no).c_str());
			}
			ComboBox_SetCurSel(hCtrl, option[optionExtSame]);
			hCtrl = GetDlgItem(hDlg, IDC_COMBO_FOLDER_SAME);
			for(no = sameQuery ; no <= sameLast ; no++) {
				ComboBox_AddString(hCtrl, GetString(IDS_OVERWRITE_QUERY + no).c_str());
			}
			ComboBox_SetCurSel(hCtrl, option[optionFolderSame]);
		}
		break;

	case WM_COMMAND:
		switch(LOWORD (wParam)) {
		case IDOK:
			{
				int regexp_mode, option[optionMax];

				regexp_mode = Button_GetCheck(GetDlgItem(hDlg, IDC_CHECK_REGEXP));

				option[optionExtMode] = ComboBox_GetCurSel(GetDlgItem(hDlg, IDC_COMBO_EXT_MODE));
				if(CheckWindows7() && option[optionExtMode] >= modeLast7) {
					option[optionExtMode] = modeLast;
				}
				option[optionExtSelect] = ComboBox_GetCurSel(GetDlgItem(hDlg, IDC_COMBO_SELECT));
				option[optionExtUnder] = Button_GetCheck(GetDlgItem(hDlg, IDC_CHECK_EXT_UNDER));
				option[optionExtAdd] = Button_GetCheck(GetDlgItem(hDlg, IDC_CHECK_EXT_ADD));
				option[optionExtSame] = ComboBox_GetCurSel(GetDlgItem(hDlg, IDC_COMBO_EXT_SAME));

				option[optionFolderMode] = ComboBox_GetCurSel(GetDlgItem(hDlg, IDC_COMBO_FOLDER_MODE));
				if(option[optionFolderMode] == 2) {
					option[optionFolderMode] = modeLast;
				}
				option[optionFolderUnder] = Button_GetCheck(GetDlgItem(hDlg, IDC_CHECK_FOLDER_UNDER));
				option[optionFolderAdd] = Button_GetCheck(GetDlgItem(hDlg, IDC_CHECK_FOLDER_ADD));
				option[optionFolderSame] = ComboBox_GetCurSel(GetDlgItem(hDlg, IDC_COMBO_FOLDER_SAME));
				// WXgɐݒ
				set_registory(regexp_mode, option);
				EndDialog(hDlg, IDOK);
			}
			break;

		case IDCANCEL:
			EndDialog(hDlg, IDCANCEL);
			break;

		}
		break;
	case WM_HELP:
		ShowHtmlHelp(hDlg, HH_HELP_CONTEXT, HLP_IDH_OPTION);
		return TRUE;
	}
	return FALSE;
}

//
//	_CAObZ[W
//
BOOL APIENTRY LogFunc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message) {
	case WM_INITDIALOG:
		{
			RECT rc;
			HWND hList = GetDlgItem(hDlg, IDC_LIST_LOG);
			LVCOLUMN lvc = { 0 };
			LV_ITEM lv = { 0 };
			double width;
			string str;
			int no;

			MouseCursorWindow(hDlg, IDOK);

			SetWindowText(hDlg, GetString((pShellExt->GetCopyFlag() == modeCopyDirect) ? IDS_RESULT_COPY : IDS_RESULT_MOVE).c_str());
			ListView_SetExtendedListViewStyle(hList, ListView_GetExtendedListViewStyle(hList) | LVS_EX_FULLROWSELECT | LVS_EX_LABELTIP);

			GetWindowRect(hList, &rc);
			width = rc.right - rc.left - GetSystemMetrics(SM_CXVSCROLL) - GetSystemMetrics(SM_CXEDGE) * 2;

			lvc.mask = LVCF_WIDTH | LVCF_TEXT;
			lvc.fmt = LVCFMT_LEFT;
			lvc.cx = static_cast<int>(width / 100 * 35);
			str = GetString((pShellExt->GetCopyFlag() == modeCopyDirect) ? IDS_RESULT_SRC_COPY : IDS_RESULT_SRC_MOVE);
			lvc.pszText = const_cast<char *>(str.c_str());
			ListView_InsertColumn(hList, 0, &lvc);
			lvc.cx = static_cast<int>(width / 100 * 35);
			str = GetString((pShellExt->GetCopyFlag() == modeCopyDirect) ? IDS_RESULT_DST_COPY : IDS_RESULT_DST_MOVE);
			lvc.pszText = const_cast<char *>(str.c_str());
			ListView_InsertColumn(hList, 1, &lvc);
			lvc.cx = static_cast<int>(width / 100 * 20);
			str = GetString(IDS_RESULT_FILE);
			lvc.pszText = const_cast<char *>(str.c_str());
			ListView_InsertColumn(hList, 2, &lvc);
			lvc.cx = static_cast<int>(width / 100 * 10);
			str = GetString(IDS_RESULT_RESULT);
			lvc.pszText = const_cast<char *>(str.c_str());
			ListView_InsertColumn(hList, 3, &lvc);

			no = 0;
			vector<COPY_RESULT> list = pShellExt->GetResultList();
			for(vector<COPY_RESULT>::iterator result = list.begin() ; result != list.end() ; result++) {
				lv.mask = LVIF_TEXT;
				lv.pszText = const_cast<char *>(result->src.c_str());
				lv.iItem = no;
				lv.iSubItem = 0;
				ListView_InsertItem(hList, &lv);
				lv.mask = LVIF_TEXT;
				lv.pszText = const_cast<char *>(result->dst.c_str());
				lv.iItem = no;
				lv.iSubItem = 1;
				ListView_SetItem(hList, &lv);
				lv.mask = LVIF_TEXT;
				lv.pszText = const_cast<char *>(result->file.c_str());
				lv.iItem = no;
				lv.iSubItem = 2;
				ListView_SetItem(hList, &lv);
				lv.mask = LVIF_TEXT;
				str = GetString((result->flag) ? IDS_RESULT_SUCCESS : IDS_RESULT_FAIL);
				lv.pszText = const_cast<char *>(str.c_str());
				lv.iItem = no;
				lv.iSubItem = 3;
				ListView_SetItem(hList, &lv);
				no++;
			}
		}
		break;

	case WM_COMMAND:
		switch(LOWORD (wParam)) {
		case IDOK:
			EndDialog(hDlg, IDOK);
			break;
		case IDCANCEL:
			EndDialog(hDlg, IDCANCEL);
			break;
		}
		break;
	case WM_HELP:
		return TRUE;
	}
	return FALSE;
}

