//
//	cpfShell e`
//
#include <vector>

enum {
	folderModeExt,
	folderModeFolder,

	folderModeMax
};
enum {
	modeCopyDirect,
	modeMove,
	modeCopy,
	modeCut,
	modeSelectOnly,
	modeLast,

	modeLast7 = modeSelectOnly,
};
enum {
	selectExt,
	selectHistory,
};

enum {
	sameQuery,
	sameNew,
	sameOverwrite,
	sameLast,
};


struct COPY_RESULT {
	std::string src;
	std::string dst;
	std::string file;
	BOOL flag;
};

STDAPI DllCanUnloadNow(void);
STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID *ppvOut);

class CShellExtClassFactory : public IClassFactory
{
protected:
	ULONG	m_cRef;

public:
	CShellExtClassFactory();
	~CShellExtClassFactory();

	//IUnknown members
	STDMETHODIMP QueryInterface(REFIID, LPVOID FAR *);
	STDMETHODIMP_(ULONG) AddRef();
	STDMETHODIMP_(ULONG) Release();

	//IClassFactory members
	STDMETHODIMP CreateInstance(LPUNKNOWN, REFIID, LPVOID FAR *);
	STDMETHODIMP LockServer(BOOL);

};
typedef CShellExtClassFactory *LPCSHELLEXTCLASSFACTORY;

class CShellExt : public IShellExtInit, 
                  public IContextMenu
{
public:
	CShellExt();
	~CShellExt();

	//IUnknown members
	STDMETHODIMP QueryInterface(REFIID, LPVOID FAR *);
	STDMETHODIMP_(ULONG) AddRef();
	STDMETHODIMP_(ULONG) Release();

	//IShell members
	STDMETHODIMP QueryContextMenu(HMENU hMenu, UINT indexMenu, UINT idCmdFirst, UINT idCmdLast, UINT uFlags);
	STDMETHODIMP InvokeCommand(LPCMINVOKECOMMANDINFO lpcmi);
	STDMETHODIMP GetCommandString(UINT idCmd, UINT uFlags, UINT FAR *reserved, LPSTR pszName, UINT cchMax);

	//IShellExtInit methods
	STDMETHODIMP Initialize(LPCITEMIDLIST pIDFolder, LPDATAOBJECT pDataObj, HKEY hKeyID);

	void SelectFile(HWND hWnd);
	void SelectFileXP(HWND hWnd);
	void FileToClipboard(HWND handle);
	void DirectCopy(HWND handle);
	void SetWildName(std::string name) { wild_name = name; }
	void SetCopyFlag(int flag) { copy_flag = flag; }
	int GetCopyFlag() { return copy_flag; }
	int GetCopyCount() { return file_list.size(); }
	void SetCancelFlag() { cancel_flag = TRUE; }
	void SetSameType(int type) { same_type = type; }
	void SetUnderType(int type) { under_type = type; }
	std::string GetFileExt(void);
	void SetComboBox(HWND h);
	void SetFolderComboBox(HWND h);
	void GetRegistry(void);
	void SetRegistry(void);
	int GetCopyMode(void) { return copy_mode[folder_mode]; }
	int GetSameMode(void) { return same_mode[folder_mode]; }
	int GetRegexpMode(void) { return regexp_mode; }
	int GetFolderMode(void) { return folder_mode; }
	int GetUnderMode(void) { return under_mode[folder_mode]; }
	int GetAddMode(void) { return add_mode[folder_mode]; }
	void SetCopyDir(std::string name) { copy_dir = name; }
	void SetCopyDirOrg(std::string name) { copy_dir_org = name; }
	std::string GetDestName() { return dst_name; }
	std::string GetFilePath() {
		if(file_path.size() == 2 && file_path[1] == ':') {
			return file_path + "\\";
		}
		return file_path;
	}
	void AddFile(std::string path);
	void AddSubFile(std::string path, BOOL flag);
	std::vector<COPY_RESULT> GetResultList() { return result_list; }
	static BOOL CALLBACK EnumChildProc(HWND hwnd, LPARAM lParam);
protected:
	ULONG m_cRef;
	LPDATAOBJECT m_pDataObj;

	std::string file_ext;			// gq
	std::string wild_name;			// ChJ[h
	std::string file_path;			// pX
	std::string copy_dir;			// Rs[tH_
	std::string copy_dir_org;		// Rs[tH_(ǉp)
	std::string dst_name;			// Rs[(ip)
	std::vector<std::string> file_list;		// t@C̈ꗗ
	std::vector<std::string> history_list[folderModeMax];	// ChJ[hꗗ
	std::vector<std::string> folder_list[folderModeMax];	// tH_ꗗ
	std::vector<COPY_RESULT> result_list;	// Rs[
	int copy_mode[folderModeMax];	// Rs[[h
	int same_mode[folderModeMax];	// ㏑Rs[[h
	int under_mode[folderModeMax];	// TutH_[h
	int add_mode[folderModeMax];	// tH_t[h
	int regexp_mode;				// K\gp邩
	int same_type;					// ㏑Rs[[h
	int under_type;					// TutH_̃t@CRs[
	int copy_flag;					// Rs[邩H
	int select_mode;				// I𕶎񃂁[h
	int file_list_size;				// t@Cꗗ̃TCY
	int copy_count;					// Rs[
	int cancel_flag;				// Rs[~tO
	int folder_mode;				// tH_[h
	HWND FindListView(HWND wnd);
};
typedef CShellExt *LPCSHELLEXT;

